import processing.serial.*;

int tileSize = 5;
int wwidth, wheight, whichWorld, wsize, mwsize, xShift, yShift;
int[] world, age1, age2, age3, age4;
int tickTimer;
IntList keys = new IntList();

Serial arduino;
boolean gotValues;
float[] pots;
boolean[] btns;

void setup() {
  gotValues = false;
  pots = new float[6];
  btns = new boolean[32];
  
  fullScreen();
  noCursor();
  colorMode(RGB, 255);
  noStroke();
  noSmooth();
  background(0);

  wwidth = width / tileSize;
  wheight = height / tileSize;
  wsize = wwidth * wheight;
  world = new int[wsize];
  age1 = new int[wsize];
  age2 = new int[wsize];
  age3 = new int[wsize];
  age4 = new int[wsize];
  whichWorld = 0;
  tickTimer = 0;
  xShift = 0;
  yShift = 0;
  
  for (int i = 0; i < wsize; i++) {
    world[i] = random(10) > 7 ? 0 : 1;
    age1[i] = world[i];
    age2[i] = 0;
    age3[i] = 0;
    age4[i] = 0;
  }
  
  // try to find the controller
  String port = "";
  while (port.equals("")) {
    for (String s : Serial.list()) {
      if (s.indexOf("usbmodem") != -1) {
        port = s;
        try {
          arduino = new Serial(this, port, 115200);
          break;
        } catch (Exception e) {
          port = ""; // keep trying
        }
      }
    }
  }
}

void draw() {
  // read settings
  updatePotentiometers();
  if (!gotValues) {
    return;
  }
  
  if (btns[5]) {
    for (int i = 0; i < wsize; i++) {
      world[i] = 0;
      age1[i] = 1 - world[i];
      age2[i] = 0;
      age3[i] = 0;
      age4[i] = 0;
    }
  }
  if (btns[4]) {
    saveFrame("life-#########.png");
  }
  
  if (btns[2]) {
    tickTimer++;
  }
  
  if (tickTimer >= (int) exp(4*(1-pots[5]))) {
    tickTimer = 0;
    simulateStep();
  } else if (btns[3]) {
    simulateStep();
  }
  
  if (keys.hasValue(UP)) yShift++;
  if (keys.hasValue(DOWN)) yShift--;
  if (keys.hasValue(RIGHT)) xShift--;
  if (keys.hasValue(LEFT)) xShift++;
    
  
  // draw
  for (int y = 0; y < wheight; y++) {
    for (int x = 0; x < wwidth; x++) {
      int index = y*wwidth + x;
      if (btns[1]) {
        if (age3[index] < age4[index]) {
          fill(getColor((int) (age4[index]*(1+0.125)+1), pots[2]*1.5));
        } else if (age2[index] < age3[index]) {
          fill(getColor((int) (age3[index]*(1+0)+1), pots[2]*1.5));
        } else {
          fill(getColor((int) (age2[index]*(1-0.125)+1), pots[2]*1.5));
        }
      } else {
        fill(getColor(age1[index], pots[2]*1.5));
      }
      int xx = ((x + xShift) % wwidth + wwidth) % wwidth;
      int yy = ((y + yShift) % wheight + wheight) % wheight;
      rect(xx * tileSize, yy * tileSize, tileSize, tileSize);
    }
  }
}

color getColor(int age, float hue) {
  if (btns[1]) {
    float b1 = exp(pots[4] * log(5000) - 1.5);
    float contrast = map(pots[3], 0, 1, 0.3, 2);
    float lightness = exp(-pow(age / b1, contrast));
    if (btns[0]) {
      lightness = 1 - lightness;
    }
    return hsl(((hue + (pots[1]-.5)*2*(lightness-.25)) % 1 + 1) % 1, 0.6, lightness);
  } else {
    if ((age == 0) == btns[0]) {
      return hsl(0, 0, 0.2);
    } else {
      return hsl(0, 0, 1);
    }
  }
}

void simulateStep() {
  for (int i = 0; i < wsize; i++) {
    world[i] <<= 1;
  }

  for (int i = 0; i < wsize; i++) {
    // count alive neighbours
    int neighbours = 0;
    if (btns[29] && (world[(i + wsize - wwidth - 1) % wsize] & 2) != 0) neighbours++;
    if (btns[30] && (world[(i + wsize - wwidth    ) % wsize] & 2) != 0) neighbours++;
    if (btns[31] && (world[(i + wsize - wwidth + 1) % wsize] & 2) != 0) neighbours++;
    if (btns[28] && (world[(i + wsize          - 1) % wsize] & 2) != 0) neighbours++;
    if (btns[24] && (world[(i + wsize          + 1) % wsize] & 2) != 0) neighbours++;
    if (btns[27] && (world[(i + wsize + wwidth - 1) % wsize] & 2) != 0) neighbours++;
    if (btns[26] && (world[(i + wsize + wwidth    ) % wsize] & 2) != 0) neighbours++;
    if (btns[25] && (world[(i + wsize + wwidth + 1) % wsize] & 2) != 0) neighbours++;
    
    // determine next state based on alive neighbour count
    boolean isAlive = false;
    boolean flipped = random(exp(pow(1-pots[0]*0.5, 4) * log(100*wwidth*wheight)) + 1) < 1;
    if (pots[0] == 0) {
      flipped = false;
    }
    boolean killed = false; //random(exp(pow(1-pots[0], 2) * log(100*wwidth*wheight)) + 1) < 1;
    if ((world[i + whichWorld] & 2) != 0) {
      if ((((btns[23] && neighbours == 0) ||
            (btns[22] && neighbours == 1) ||
            (btns[21] && neighbours == 2) ||
            (btns[20] && neighbours == 3) ||
            (btns[19] && neighbours == 4) ||
            (btns[18] && neighbours == 5) ||
            (btns[17] && neighbours == 6) ||
            (btns[16] && neighbours == 7) ||
            (btns[6] && neighbours == 8)
            ) != flipped) && !killed)
        isAlive = true;
    } else {
      if ((((btns[15] && neighbours == 0) ||
            (btns[14] && neighbours == 1) ||
            (btns[13] && neighbours == 2) ||
            (btns[12] && neighbours == 3) ||
            (btns[11] && neighbours == 4) ||
            (btns[10] && neighbours == 5) ||
            (btns[9] && neighbours == 6) ||
            (btns[8] && neighbours == 7) ||
            (btns[7] && neighbours == 8)
            ) != flipped) && !killed)
        isAlive = true;
    }
    world[i] += isAlive ? 1 : 0;
    
    if ((world[i] & 1) == 1) {
      age1[i] = 0;
    } else {
      age1[i]++;
    }
    
    if ((world[i] & 3) == 2 ||
        (world[i] & 3) == 1) {
      age2[i] = 0;
    } else {
      age2[i]++;
    }
    
    // magic, these just look nice
    if ((world[i] & 7) == 6 ||
        (world[i] & 7) == 3 ||
        (world[i] & 15) == 7 ||
        (world[i] & 31) == 15 ||
        (world[i] & 63) == 31) {
      age3[i] = 0;
    } else {
      age3[i]++;
    }
    
    if ((world[i] & 15) == 14 ||
        (world[i] & 15) == 7 ||
        (world[i] & 31) == 15 ||
        (world[i] & 63) == 31 ||
        (world[i] & 127) == 63) {
      age4[i] = 0;
    } else {
      age4[i]++;
    }
  }
}

color hsl(float hue, float saturation, float luma) {
  float r, g, b;
  float p, q;
  if (saturation == 0) {
    r = g = b = luma;
  } else {
    if (luma < 0.5) {
      q = luma * (1 + saturation);
    } else {
      q = (luma + saturation) - (saturation * luma);
    }
    p = 2 * luma - q;
    r = hue2rgb(p, q, hue + 1./3);
    g = hue2rgb(p, q, hue);
    b = hue2rgb(p, q, hue - 1./3);
  }
  float m = luma - (.21*r + .72*g + .07*b);
  return color((int) (255 * (r+m)), (int) (255 * (g+m)), (int) (255 * (b+m)));
}

float hue2rgb(float p, float q, float t) {
  if (t < 0) t += 1;
  if (t > 1) t -= 1;
  if (t < 1./6)
    return p + (q - p) * 6 * t;
  if (t < 1./2)
    return q;
  if (t < 2./3)
    return p + (q - p) * (2./3 - t) * 6;
  return p;
}

void keyPressed() {
  keys.append(keyCode);
}

void keyReleased() {
  // could not find a method to do this
  int index = 0;
  while (index < keys.size()) {
    if (keys.get(index) == keyCode) {
      keys.remove(index);
    } else {
      index++;
    }
  }
}

void updatePotentiometers() {
  arduino.write('h'); // tell arduino we can read the next values
  if (arduino.available() > 0) {
    String buffer = readLine(arduino);
    String numbers = buffer.split(";")[0];
    String booleans = buffer.split(";")[1];
    try {
      for (int i = 0; i < 6; i++) {
        pots[i] = parseValue(numbers.split(",")[i]);
      }
      gotValues = true;
    } catch (NumberFormatException e) {}
    for (int i = 0; i < 32; i++) {
      btns[i] = booleans.split(",")[i].equals("1");
    }
  }
}

float parseValue(String s) {
  float number = (float) Integer.parseInt(s);
  return map(number, 0, 1023, 0, 1);
}

String readLine(Serial device) {
  String buffer = "";
  while (true) {
    char inChar = device.readChar();
    if (inChar == '\n') {
      break;
    }
    if (inChar == '\r') {
      continue;
    }
    buffer += inChar;
  }
  return buffer;
}